import pandas as pd
import numpy as np


def handle_data(
    df: pd.DataFrame, date_col: str | None = None, log_cols: list[str] = []
) -> pd.DataFrame:
    df = df.copy()

    if date_col:
        df[date_col] = pd.to_datetime(df[date_col])

    cat_cols = df.select_dtypes(include=["category", "object"]).columns.tolist()

    # Ensure dtype is category for accessing .cat and for LGBM training
    df[cat_cols] = df[cat_cols].astype("category")

    # df[cat_cols] = df[cat_cols].fillna("Nan")

    for col in cat_cols:
        df[col] = df[col].cat.add_categories("Nan")
        df[col] = df[col].fillna("Nan")

    df[log_cols] = np.log1p(df[log_cols])

    return df



import pandas as pd
from sklearn.model_selection import train_test_split
from sklearn.metrics import (
    mean_squared_error,
    mean_absolute_error,
    accuracy_score,
    f1_score,
)

import numpy as np
import warnings
import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix, ConfusionMatrixDisplay

# import seaborn as sns

# Models
from catboost import CatBoostRegressor, CatBoostClassifier, Pool
from lightgbm import LGBMClassifier, LGBMRegressor
import lightgbm as lgb

# import xgboost as xgb

from typing import Literal
from .types import *

all_models = {
    "catboost": {
        "regression": CatBoostRegressor,
        "classification": CatBoostClassifier,
    },
    "lightgbm": {
        "regression": LGBMRegressor,
        "classification": LGBMClassifier,
    },
}

param_map = {
    "catboost": {
        "depth": "depth",
        "learning_rate": "learning_rate",
        "iterations": "iterations",
        "loss_function": "loss_function",
        "metric": "eval_metric",
        "reg_2": "l2_leaf_reg",
        "border_count": "border_count",
        "subsample": "subsample",
        "device": "task_type",
        "seed": "random_seed",
    },
    "lightgbm": {
        "depth": "max_depth",
        "learning_rate": "learning_rate",
        "iterations": "num_iterations",
        "loss_function": "objective",
        "metric": "metric",
        "reg_2": "lambda_l2",
        # 'border_count': 'max_bin',
        "subsample": "bagging_fraction",
        "device": "device",
        "seed": "seed",
    },
}


def map_params(model_type: str, user_params: dict) -> dict:
    model_param_map = param_map[model_type]

    mapped_params = {}
    for key, value in user_params.items():
        if key == "loss_function":
            value = map_loss_functions(value, model_type)

        mapped_key = model_param_map.get(key)
        if mapped_key:
            mapped_params[mapped_key] = value
        else:
            warnings.warn(f"Parameter {key} not recognized, and will be ignored")

    return mapped_params


loss_fn_map = {
    "catboost": {
        "RMSE": "RMSE",
        "MAE": "MAE",
        "binary": "Logloss",
        "cross_entropy": "CrossEntropy",
        "multi_class": "MultiClass",
        "MAPE": "MAPE",
        "quantile": "Quantile",
    },
    "lightgbm": {
        "RMSE": "rmse",
        "MAE": "l1",
        "binary": "binary",
        "cross_entropy": "cross_entropy",
        "multi_class": "multiclass",
        "MAPE": "mape",
        "quantile": "quantile",
    },
}


def map_loss_functions(loss_fn, model_type):
    model_loss_map = loss_fn_map[model_type]

    mapped_loss = model_loss_map.get(loss_fn)

    if mapped_loss:
        return mapped_loss
    else:
        warnings.warn(
            f"mapping for loss function: {loss_fn} not found, will be used as is"
        )
        return loss_fn


default_params = {
    "iterations": 1000,
    "learning_rate": 1e-2,
    "loss_function": "RMSE",
    "device": "CPU",
}


def get_model(
    model_type: ModelNameType = "catboost",
    task: TaskType = "regression",
    params: dict = default_params,
):
    model_category = all_models.get(model_type)

    if model_category is None:
        raise ValueError(
            f"Model: {model_type} is not recognized, use one of following: {', '.join(param_map.keys())}"
        )

    model_instance = model_category.get(task)
    if model_instance is None:
        raise ValueError(
            f"Model: {model_type} is not recognized, use one of following: {', '.join(model_category.keys())}"
        )

    mapped_params = map_params(model_type, params)

    if model_type == "lightgbm":
        # Remove Info Logs for LGBM
        mapped_params = {**mapped_params, "verbosity": -1}

    model = model_instance(**mapped_params)

    return model


class ModelWrapper:

    def __init__(
        self,
        model,
        model_type: ModelNameType,
        task: TaskType,
        X_valid: pd.DataFrame,
        y_valid: pd.Series,
        valid_preds: np.ndarray,
        eval_score: float,
    ):
        self.model = model
        self.model_type = model_type
        self.task = task
        self.X_valid = X_valid
        self.y_valid = y_valid
        self.valid_preds = valid_preds
        self.eval_score = eval_score

    def plot_importance(
        self, max_num_features: int = 20, figsize: tuple[float, float] = (10, 6)
    ):
        if self.model_type == "catboost":
            importance = self.model.get_feature_importance()
        elif self.model_type == "lightgbm":
            importance = self.model.feature_importances_
        else:
            raise ValueError(f"{self.model_type} plot_importance is not supported")

        feature_names = self.X_valid.columns.tolist()

        importance_df = pd.DataFrame(
            {"feature": feature_names, "importance": importance}
        ).sort_values(by="importance", ascending=True)

        importance_df = importance_df.head(max_num_features)

        plt.figure(figsize=figsize)
        plt.barh(importance_df["feature"], importance_df["importance"])
        plt.xlabel("Importance")
        plt.title("Feature Importance")
        plt.show()

    def confusion_matrix(self):
        y_valid, valid_preds = self.y_valid, self.valid_preds
        cm = confusion_matrix(y_valid, valid_preds)

        disp = ConfusionMatrixDisplay(confusion_matrix=cm)
        disp.plot(values_format="d")
        plt.title("Confusion Matrix")
        plt.show()

    def __getattr__(self, attr):
        return getattr(self.model, attr)


metric_map = {
    "mse": mean_squared_error,
    "mae": mean_absolute_error,
    "accuracy": accuracy_score,
    "f1": f1_score,
}


def train_model(
    model_type: ModelNameType = "catboost",
    task: TaskType = "regression",
    params: dict = default_params,
    train_data: pd.DataFrame | None = None,
    valid_data: pd.DataFrame | None = None,
    target_col: str = "target",
    metric: Literal["mse", "mae", "accuracy", "f1"] | None = None,
    verbose: int = 100,
    early_stop: int = 500,
    random_state: int | None = 42,
    test_size: float | None = 0.2,
):
    model = get_model(model_type, task, params)

    if train_data is None:
        raise ValueError("train_data is not found")

    if train_data is not None and valid_data is not None:
        X_train = train_data.drop(columns=target_col)
        y_train = train_data[target_col]
        X_valid = valid_data.drop(columns=target_col)
        y_valid = valid_data[target_col]

    else:
        print(
            "valid_data is not explicitly passed, hence performing random split on train_data"
        )
        X = train_data.drop(columns=[target_col])
        y = train_data[target_col]
        X_train, X_valid, y_train, y_valid = train_test_split(
            X, y, test_size=test_size, random_state=random_state
        )

    cat_cols = X_train.select_dtypes(include=["category", "object"]).columns.tolist()

    if model_type == "catboost":
        train_pool = Pool(X_train, y_train, cat_features=cat_cols)
        valid_pool = Pool(X_valid, y_valid, cat_features=cat_cols)

        model.fit(
            train_pool,
            eval_set=[train_pool, valid_pool],
            use_best_model=True,
            verbose=verbose,
            early_stopping_rounds=early_stop,
        )

    elif model_type == "lightgbm":
        model.fit(
            X_train,
            y_train,
            eval_set=[(X_train, y_train), (X_valid, y_valid)],
            callbacks=[
                lgb.early_stopping(stopping_rounds=early_stop),
                lgb.log_evaluation(period=verbose),
            ],
        )

    if task == "classification":
        # Warn if invalid metric provided and set classification default
        if metric in ("mse", "mae"):
            warnings.warn(
                f"{metric} is not a valid metric, using default metric 'accuracy' instead"
            )
            metric = "accuracy"

        # Set classification default metric
        if metric is None:
            metric = "accuracy"

    else:
        # Warn if invalid metric provided and set regression default
        if metric in ("accuracy", "f1"):
            warnings.warn(
                f"{metric} is not a valid metric, using default metric 'mse' instead"
            )
            metric = "mse"

        # Set regression default
        if metric is None:
            metric = "mse"

    metric_fn = metric_map[metric]

    valid_preds = model.predict(X_valid)
    eval_score = metric_fn(y_valid, valid_preds)

    print(f"\nValidation {metric} score: {eval_score}")

    model = ModelWrapper(
        model, model_type, task, X_valid, y_valid, valid_preds, eval_score
    )

    return model



from sklearn.model_selection import KFold, StratifiedKFold
import pandas as pd
import numpy as np
import warnings

from .model import train_model, default_params

from typing import Literal
from .types import *


def fold_train(
    model_type: ModelNameType = "catboost",
    task: TaskType = "regression",
    params: dict = default_params,
    data: pd.DataFrame | None = None,
    target_col: str = "target",
    n_splits: int = 5,
    metric: Literal["mse", "mae", "accuracy", "f1"] | None = None,
    verbose: int = 100,
    early_stop: int = 500,
    random_state: int | None = 42,
):
    if data is None:
        raise ValueError("dataset not found")

    # reset index to prevent indexing errors
    data = data.copy()
    data = data.reset_index(drop=True)

    if task == "classification":
        X = data.drop(columns=[target_col])
        y = data[target_col]
        skf = StratifiedKFold(n_splits=5, shuffle=True, random_state=42)
        split = skf.split(X, y)

    else:
        kf = KFold(n_splits=n_splits, shuffle=True, random_state=42)
        split = kf.split(data)

    model_list = []
    eval_scores = []

    for fold_num, (train_idx, valid_idx) in enumerate(split):
        print(f"Running Fold: {fold_num + 1}")

        train_df = data.loc[train_idx].copy()
        valid_df = data.loc[valid_idx].copy()

        model = train_model(
            model_type=model_type,
            task=task,
            params=params,
            train_data=train_df,
            valid_data=valid_df,
            target_col=target_col,
            metric=metric,
            verbose=verbose,
            early_stop=early_stop,
            random_state=random_state,
        )

        model_list.append(model)
        eval_scores.append(model.eval_score)

    mean_score = np.array(eval_scores).mean()

    print(f"\nMean validation score: {mean_score}")

    return model_list


def get_fold_preds(models, test_df: pd.DataFrame) -> np.ndarray:
    all_preds = []

    for model in models:
        preds = model.predict(test_df)
        all_preds.append(preds)

    mean_preds = np.array(all_preds).mean(axis=0)

    return mean_preds


def get_ensemble_preds(preds_list, weights: list[float] | None = None):
    if weights is not None and len(preds_list) != len(weights):
        raise ValueError(f"number of preds_list and number of weights do not match")

    preds_list = np.array(preds_list)
    weighted_preds = np.average(preds_list, weights=weights, axis=0)

    return weighted_preds



import pandas as pd
import numpy as np
import warnings
from itertools import combinations
from sklearn.model_selection import KFold
from typing import Any

# ignoring this warning since it triggers too many times, cluttering output, and same warning provided in combine_features already
warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)


def expand_date(df: pd.DataFrame, date_col: str = "date_time"):
    df = df.copy()

    if date_col not in df.columns:
        ValueError(f"Column: {date_col} is not found in df")

    if not pd.api.types.is_datetime64_any_dtype(df[date_col]):
        warnings.warn(
            f"Column: {date_col} is not a datetime type, automatically converting datetime"
        )
        df[date_col] = pd.to_datetime(df[date_col])

    df["year"] = df[date_col].dt.year
    df["month"] = df[date_col].dt.month
    df["day"] = df[date_col].dt.day
    df["weekday"] = df[date_col].dt.weekday
    df["is_weekend"] = (df[date_col].dt.weekday) > 4
    df["week_of_year"] = df[date_col].dt.isocalendar().week

    return df


def combine_features(
    df: pd.DataFrame,
    target_col: str | None = None,
    num_features: list[str] | None = None,
    cat_features: list[str] | None = None,
    # methods=['divide', 'multiply']
    safety_check: bool = True,
) -> pd.DataFrame:

    # reset index to prevent indexing errors
    df = df.copy()
    df = df.reset_index(drop=True)

    y = None

    if target_col is None:
        warnings.warn("target_col not specified, this may lead to data leakage")
    else:
        y = df[target_col]
        df = df.drop(columns=target_col)

    if not num_features:
        warnings.warn(
            "num_features not provided, using all numerical features available, this may lead to poor performance"
        )
        num_features = df.select_dtypes(include=["number"]).columns.tolist()

    if not cat_features:
        warnings.warn(
            "cat_features not provided, using all categorical features available, this may lead to poor performance"
        )
        cat_features = df.select_dtypes(include=["category", "object"]).columns.tolist()

    if (len(num_features) > 50 or len(cat_features) > 50) and safety_check:
        raise ValueError(
            "More than 50 features selected, this could lead to extremely high memory usage or slow performance. Pass safety_check = False to override this safe check"
        )

    for col1, col2 in combinations(num_features, 2):
        df[f"{col1}_times_{col2}"] = (df[col1] * df[col2]).astype(float)
        df[f"{col1}_div_{col2}"] = (df[col1] / df[col2].replace(0, np.nan)).astype(
            float
        )

    for col1, col2 in combinations(cat_features, 2):
        df[f"{col1}_{col2}"] = (
            df[col1].astype(str) + "_" + df[col2].astype(str)
        ).astype("category")

    if y is not None:
        df[target_col] = y.copy()

    return df


def bin_column(df: pd.DataFrame, col: str, bins: int = 4, labels=None) -> pd.DataFrame:
    df[f"{col}_binned"] = pd.cut(df[col], bins=bins, labels=labels)
    return df


def target_encode_test(
    train_df: pd.DataFrame,
    test_df: pd.DataFrame,
    target_col: str = "target",
    cols: list[str] | None = None,
) -> pd.DataFrame:

    # reset index to prevent indexing errors
    train_df = train_df.copy()
    train_df = train_df.reset_index(drop=True)

    test_df = test_df.copy()
    test_df = test_df.reset_index(drop=True)

    if not pd.api.types.is_numeric_dtype(train_df[target_col]):
        raise TypeError(f"{target_col} is not numeric")

    if not cols:
        raise ValueError(f"cols not provided")

    for col in cols:
        if train_df[col].dtype.name not in ["object", "category"]:
            # warnings.warn(f"{col} is numerical, therefore will be binned")
            train_df = bin_column(train_df, col)
            test_df = bin_column(test_df, col)

        # agg_list = [np.mean, np.median, np.max, np.min]
        # using Any type as work around for .agg function-string conflict
        agg_list: Any = ["mean", "median", "max", "min"]
        group_df = (
            train_df.groupby(col, observed=False)[target_col]
            .agg(agg_list)
            .reset_index()
        )
        group_df.columns = [col] + [f"{col}_te_{name}" for name in agg_list]

        test_df = pd.merge(test_df, group_df, on=col, how="left")

        # Remove bin features added from bin_column, ignore key error in case _binned columns don't exist
        test_df.drop(columns=[f"{col}_binned"], inplace=True, errors="ignore")

    return test_df


def target_encode(
    df: pd.DataFrame, target_col: str | None = None, cols: list[str] | None = None
) -> pd.DataFrame:

    # reset index to prevent indexing errors
    df = df.copy()
    df = df.reset_index(drop=True)

    if not target_col:
        raise ValueError(f"target_col not provided")

    if not cols:
        raise ValueError(f"cols not provided")

    kf = KFold(n_splits=5, shuffle=True, random_state=42)

    te_df = pd.DataFrame()

    for train_idx, valid_idx in kf.split(df):
        train_df = df.loc[train_idx].copy()
        valid_df = df.loc[valid_idx].copy()

        te_valid_df = target_encode_test(train_df, valid_df, target_col, cols)
        te_df = pd.concat([te_df, te_valid_df], ignore_index=True)

    return te_df



from typing import Literal, Union
from catboost import CatBoostRegressor, CatBoostClassifier
from lightgbm import LGBMRegressor, LGBMClassifier

# from xgboost import XGBRegressor, XGBClassifier

ModelNameType = Literal["catboost", "lightgbm", "xgboost"]
# ModelType = (
#     CatBoostRegressor
#     | CatBoostClassifier
#     | LGBMRegressor
#     | LGBMClassifier
#     # | XGBRegressor
#     # | XGBClassifier
# )


TaskType = Literal["regression", "classification"]
